/**
 * Module dependencies
 */

var EventEmitter  = require('events').EventEmitter;
var extend        = require('seed-extend');
var each          = require('seed-each');
var chronicler    = require('chronicler');
var drtp          = require('driver-rtp');
var Vue           = require('vue');
var Router        = require('seed-router');
var Store         = require('seed-store');
var VueStore      = require('seed-store/plugins/vue');

// Vue.config.debug = true;

/**
 * Constructor
 *
 * @param {Object} options
 *   @param {String} name, api namespace
 *   @param {Object} handshake, handshake payload
 *   @param {String} el, anchor element for the root application
 *   @param {Object} bundle, components bundle
 *     @param {Object} components
 *     @param {Object} directives
 *     @param {Object} filters
 *   @param {String} template, view layout
 *   @param {Array} mixins
 *   @param {Object} data, default view data, optional
 *   @param {String} currentView, default view, optional
 *   @param {Object} config, global vuejs config, optional
 */

function VueEngine(options) {
  if (!(this instanceof VueEngine)) {
    return new VueEngine(options);
  }
  
  // Vue.config(options.config || {
  //   // silent: true TODO: uncomment in production
  // });
  
  var self = this;
  
  // setup vue
  options.bundle = options.bundle || {};
  
  options.mixins = options.mixins || [];
  
  // emitter
  options.mixins.push({
    methods: {
      emit: function(type, data, fn) {
        var emitter = this.$root || this;
        emitter.$emit('ve', {
          type: type,
          data: data,
          fn: fn
        });
      }
    }
  });
  
  // setup store
  var store = this.store = new Store();
  store.use(VueStore);
  options.mixins.push(store.mixin);
  
  each(options.bundle.components || {}, function(view, name) {
    view.mixins = view.mixins || [];
    options.mixins.forEach(function(mixin) {
      view.mixins.push(mixin);
    });
    Vue.component(name, view);
  });
  
  each(options.bundle.directives || {}, function(view, name) {
    Vue.directive(name, view);
  });
  
  each(options.bundle.filters || {}, function(view, name) {
    Vue.filter(name, view);
  });
  
  each(options.bundle.observers || {}, function(obs, name) {
    obs(self);
  });
  
  var data = options.data || {};
  data.currentView = options.currentView || '';
  
  this.view = new Vue({ 
    mixins:   options.mixins,
    cursors:  options.cursors || {},
    props:    options.props || [],
    el:       options.el || '#root',
    template: options.template,
    data:     data
  });
  
  // setup event comms VueEngine <-> Vuejs
  this.view.$on('ve', function(ev) {
    self.emit(ev.type, ev.data, ev.fn);
  });
  
  self.on('set', function(key, value) {
    self.view.$set(key, value);
  });
  
  self.on('route', function(uri, query) {
    Vue.nextTick(function() {
      self.view.$emit('route', uri, query);
    });
  });
  
  self.on('render', function(view, locals, opts) {
    self.view.$data.currentView = view;
    Vue.nextTick(function() {
      if (self.view.$.content) {
        extend(self.view.$.content.$data, locals);
      } else {
        console.log(view, 'not found');
      }
    });
  });
  
  // setup router
  this._hash = chronicler();
  this._hash.on('change', this._route.bind(this));
  this._router = new Router();
  this._router.add(function(ctx, next) {
    self.emit('route', ctx.uri, ctx.query);
    next();
  });
  
  // setup remote api
  this._api = {}; // placeholder
  this._router.add(function(ctx, next) {
    ctx.api = self._api;
    next();
  });
  drtp({
    namespace: options.name,
    handshake: options.handshake || {}
  }, function(api, client) {
    window.api = self._api = api;
    self._client = client;
    self.emit('ready');
  });
  
  // setup app qualities
  this._router.add(function(ctx, next) {
    ctx.title = function(str) {
      window.document.title = str;
      return ctx;
    }
    ctx.redirect = function(uri, cb) {
      // TODO: support object
      window.location.hash = uri;
      self._router.set(uri, cb);
      return ctx;
    }
    ctx.rewrite = function(uri) {
      self._rewrite(uri);
      return ctx;
    }
    ctx.render = function(template, locals, options) {
      self.emit('render', template, locals, options);
      return ctx;
    }
    ctx.set = function(key, value) {
      self.emit('set', key, value);
      return ctx;
    }
    ctx.get = function(key) {
      return self.view.$data[key];
    }
    next();
  });
  
  EventEmitter.call(this);
}

/**
 * Extends eventemitter
 */

extend(VueEngine.prototype, EventEmitter.prototype);


VueEngine.prototype._route = function(hash) {  
  
  if (this._rewriting) {
    this._rewriting = false;
    return;
  }
  
  hash = hash || window.location.hash;
  
  if (hash.indexOf('#') === 0) {
    hash = hash.substring(1, hash.length);
  }
  
  if (hash.indexOf('/') !== 0) {
    hash = '/' + hash;
    // TODO: overwrite window location hash to /#/...?{query}
    window.location.hash = hash;
    return;
  }
  
  this._router.set(hash, function(error, result) {
    // TODO: do something here
    console.log(hash, error, result);
  });
}

VueEngine.prototype._rewrite = function(hash) {
  this._rewriting = true;
  this._hash.goTo(hash);
}

VueEngine.prototype.start = function() {
  if (this._ready) { 
    this._hash.start();
  } else {
    this.on('ready', function() {
      this._ready = true;
      this._hash.start();
    }.bind(this));
  }
}

VueEngine.prototype.stop = function() {
  this._hash.stop();
}

VueEngine.prototype.router = function() {
  return this._router;
}

VueEngine.prototype.get = function(key) {
  return this.view.$data[key];
}

VueEngine.prototype.set = function(key, value) {
  this.emit('set', key, value);
  return this;
}

VueEngine.prototype.redirect = function(uri, cb) {
  // TODO: support object
  window.location.hash = uri;
  this._router.set(uri, cb);
  return this;
}

/**
 * Expose
 */

exports = module.exports = VueEngine;
exports.Router = Router;